#include <ExtendedLog/EntryType.hpp>
#include <ExtendedLog/LogWriter.hpp>
#include <QByteArray>
#include <QCloseEvent>
#include <QDateTime>
#include <QDir>
#include <QFile>
#include <QFileInfoList>
#include <QMessageBox>
#include <QStringList>

#include "exif.hpp"
#include "rename.hpp"

Rename::Rename(QString path, int minutes, int hours, ExtendedLog::LogWriter *logWriter, QObject *parent) :
	QObject{parent},
	dir(path),
	minutes(minutes),
	hours(hours),
	logWriter(logWriter)
{

}

Rename::~Rename()
{
	delete logWriter;
}

void Rename::rename()
{
	if (logWriter == nullptr)
	{
		QString errorMessage = tr("No log initialized.");

		/*
		 * We have to signal that somethig goes wrong. Of course when
		 * `logWriter` is a nullpointer must not use it. Luckily we emit an
		 * `error` signal that is caught by our progress windows.
		 */
		//logWriter->writeEntry(extLog::EntryType::ERROR, errorMessage);

		emit error(errorMessage);
		return;
	}

	if (dir == "")
	{
		QString errorMessage = tr("No path selected!");
		logWriter->writeEntry(extLog::EntryType::ERROR, errorMessage);
		emit error(errorMessage);
		return;
	}

	logWriter->writeEntry(extLog::EntryType::INFORMATION, tr("%1 selected as path.").arg(dir));

	QDir working_dir(dir);
	working_dir.setNameFilters(filters);
	working_dir.setFilter(QDir::Files);

	auto files = working_dir.entryInfoList();

	int numberOfFiles = files.count();
	int currentFiles = 0;
	int errorCount = 0;

	for (auto &&file : files)
	{
		QString fullFileName = file.canonicalFilePath(); // full filename (incl. path)
		QString fileName = file.fileName(); // filename without path

		currentFiles++;

		// Logging
		logWriter->writeEntry(extLog::EntryType::INFORMATION, tr("%1 selected.").arg(fileName));

		// check for JPEG-file not necessary because of filtering !?!

		QString message = "";
		QString newName = "";
		QString newFile = "";

		QFile photoFile(fullFileName);
		if (photoFile.open(QIODevice::ReadWrite))
		{
			QByteArray data = photoFile.readAll();
			easyexif::EXIFInfo exifInfo;
			if (int code = exifInfo.parseFrom((unsigned char*)data.data(), data.size()))
			{
                logWriter->writeEntry(
                                extLog::EntryType::ERROR,
                                tr("Error parsing EXIF-Data from %1: Code %2").arg(fullFileName).arg(code)
                           );

                message = tr("Error parsing EXIF-Data: Code %1").arg(code);
			}
			else
			{
                std::string exifDateTime = exifInfo.DateTimeOriginal != "" ? exifInfo.DateTimeOriginal : exifInfo.DateTime;

                // only rename iff we have date and time information in the file
                if (exifDateTime != "")
                {
                    QString dateTimeString = correctDateTimeFormat((QString::fromStdString(exifDateTime)));
                    QDateTime dateTime = QDateTime::fromString(dateTimeString, Qt::ISODate);

                    newFile = newFilename(dateTime, hours, minutes);

					QString pathSeperator = dir.endsWith('/')? "" : "/";
                    newName = dir + pathSeperator + newFile;
                    logWriter->writeEntry(
                                extLog::EntryType::INFORMATION,
                                tr("Date %1 found.").arg(dateTime.toString("dd.MM.yyyy hh:mm:ss"))
                            );

                    // renaming
                    int count = 1;
                    QString countName = newName;
                    if (newName == fullFileName)
                    {
                        message = tr("File %1 has already the correct name.").arg(fileName);
                        logWriter->writeEntry(extLog::EntryType::INFORMATION, message);
                    }
                    else
                    {
                        while (QFile::exists(countName))
                        {
                            QString countNameTmp = newName.left(newName.length() - 4) + "_" + QString::number(++count) + ".jpg";
							message = tr("File %1 already exists. Trying %2 as new filename.").arg(countName, countNameTmp);
                            logWriter->writeEntry(extLog::EntryType::WARNING, message);
                            countName = countNameTmp;
                        }
                        if (photoFile.rename(countName))
                        {
                            message = fileName + " -> " + newFile;
                            logWriter->writeEntry(
                                        extLog::EntryType::INFORMATION,
										tr("File %1 renamed to %2.").arg(fileName, newName)
                                    );
                        }
                        else
                        {
                            message = tr("Rename %1 failed.").arg(fileName);
                            logWriter->writeEntry(extLog::EntryType::INFORMATION, message);
                            errorCount++;
                        }
                    }

                }
                // no information -> we do nothing
                else
                {
                    message = tr("No EXIF-Data found.");
                }
            }
		}

		// renaming is this realy needed here?!?
		/*int count = 1;
		QString countName = newName;
		if (newName == fullFileName)
		{
			message = tr("File %1 has already the correct name.").arg(fileName);
			logWriter->writeEntry(extLog::EntryType::INFORMATION, message);
		}
		else
		{
			while (QFile::exists(countName))
			{
				QString countNameTmp = newName.left(newName.length() - 4) + "_" + QString::number(++count) + ".jpg";
				message = tr("File %1 already exists. Trying %2 as new filename.").arg(countName, countNameTmp);
				logWriter->writeEntry(extLog::EntryType::WARNING, message);
				countName = countNameTmp;
			}
			if (photoFile.rename(countName))
			{
				message = fileName + " -> " + newFile;
				logWriter->writeEntry(
							extLog::EntryType::INFORMATION,
							tr("File %1 renamed to %2.").arg(fileName, newName)
						);
			}
			else
			{
				message = tr("Rename %1 failed.").arg(fileName);
				logWriter->writeEntry(extLog::EntryType::INFORMATION, message);
				errorCount++;
			}
		}*/

		QStringList userState;
		userState << QString::number(currentFiles) << QString::number(numberOfFiles) << message;
		emit progressChanged(getPercent(currentFiles, numberOfFiles), userState);
	}
	logWriter->writeEntry(extLog::EntryType::INFORMATION, tr("The task was completed with %n error(s).", "", errorCount));
	logWriter->close();
	emit finished();
}

QString Rename::newFilename(QDateTime datetime, int hr_plus, int min_plus)
{
	QString new_filename = "";
	qlonglong adjusting = (hr_plus * 3600) + (min_plus * 60);
	QDateTime datetime_adjusted = datetime.addSecs(adjusting);
	new_filename += QString::number(datetime_adjusted.date().year()); // for the year we do not need to add padding zeros
	new_filename += addZero(datetime_adjusted.date().month());
	new_filename += addZero(datetime_adjusted.date().day());
	new_filename += "_";
	new_filename += addZero(datetime_adjusted.time().hour());
	new_filename += addZero(datetime_adjusted.time().minute());
	new_filename += addZero(datetime_adjusted.time().second());
	new_filename += ".jpg";
	return new_filename;
}

QString Rename::addZero(int value)
{
	QString stringValue = QString::number(value);
	return (value >= 0 && value < 10) ? QString::number(0) + stringValue : stringValue;
}

int Rename::getPercent(double curr, double numberOfFiles)
{
	double percent = curr / numberOfFiles * 100;
	return (int)percent;
}

QString Rename::correctDateTimeFormat(QString fromExifLibrary)
{
	QString result = fromExifLibrary;
    if (result != "")
    {
        result.replace(4, 1, '-');
        result.replace(7, 1, '-');
        result.replace(10, 1, 'T');
    }
	return result;
}

